VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "KMLPlacemark"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = False
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Compare Database
Option Explicit

'****************KML PLACEMARK OBJECT v1.0***********IRONFELIX717

'== ABOUT =======================================================
'   https://www.access-programmers.co.uk    Author: @ironfelix717
'   11-7-2022

'   1-10-23 - Fixed 2 minor bugs in .Build()
'             Debug was being logged when PATH start/end didn't match - not necessaru
'             Debug incorrectly stated end points must match for "path" when should've been "polygon"
'================================================================

'== DEPENDENCIES ================================================
'   MICROSOFT XML V6.0 (MSXML)
'   MICROSOFT SCRIPTING RUNTIME
'================================================================

'== NOTES =======================================================
'   This is a placemark object class for building/reading KML
'   placemark elements. The object supports the following placemark
'   types:  Point, Path, Polygon, Image Overlay. Screen Overlays
'   and other nontypical elements are not supported.
'   See docs for usage: https://www.access-programmers.co.uk
'                                                      God Bless
'================================================================

'== CREDITS / SUPPORTING DOCS ===================================
'   https://stackoverflow.com - great resource for MSXML2 related info
'   https://learn.microsoft.com/en-us/previous-versions/troubleshoot/msxml/list-of-xml-parser-versions
'   https://developers.google.com/kml/documentation/kml_tut
'   https://developers.google.com/kml/documentation


'PUBLIC
Dim SQL                 As String
Dim rs                  As DAO.Recordset
Dim db                  As Database
Dim fso                 As New FileSystemObject

'PROPS
Dim vDebugString        As String
Dim vName               As String
Dim vDescription        As String
Dim vStyleURL           As String
Dim vType               As KMLPlacemarkTypes
Dim vCoordinates        As String
Dim vPhotoColor         As String
Dim vPhotoIMG           As String 'href path to image
Dim vPhotoRotate        As Double 'degrees
Dim vAltitude           As Long  'metersonly whole number
Dim vAltitudeMode       As KMLAltitudeModes
Dim vExtended           As Boolean
Dim vInvertCoordinates  As Boolean
'ENUM
Enum KMLPlacemarkTypes
kPoint = 1
kPath = 2
kPolygon = 3
kPhotoOverlay = 4
End Enum

Enum KMLAltitudeModes
kClampToGround = 1
kRelativeToGround = 2
kAbsolute = 3
End Enum


Property Get DebugString() As String
    DebugString = vDebugString
End Property
    
Property Let PlacemarkName(val As String)
    vName = val
End Property
Property Get PlacemarkName() As String
    PlacemarkName = vName
End Property


Property Let PlacemarkType(val As KMLPlacemarkTypes)
    vType = val
End Property
Property Get PlacemarkType() As KMLPlacemarkTypes
    PlacemarkType = vType
End Property


Property Let Description(val As String)
    vDescription = val
End Property
Property Get Description() As String
    Description = vDescription
End Property

Property Let InvertCoordinates(val As Boolean)
    'KML requires Long/Lat - Most people use Lat first. InvertCoordinates = true if passing Lat/Long
    'ignored if photo overlay
    vInvertCoordinates = val
End Property

Property Let Coordinates(val As String)
'LAT/LONG CORDINATES ONLY - DO NOT INCLUDE ALTITUDE (use .FormatCoords() to and deformat if necessary)
'ORDER OF EACH CORDINATES MATTERS
'Also see '.InvertCoordinates
'IF DOING OVERLAY - THEN FORMAT MUST BE  N.S.E.W AND BE LAT, LAT, LONG, LONG
    vCoordinates = val
End Property
Property Get Coordinates() As String
    Coordinates = vCoordinates
End Property


Property Let StyleURL(val As String)
    vStyleURL = val
End Property
Property Get StyleURL() As String
    StyleURL = vStyleURL
End Property

Property Let PhotoColor(val As String)
    vPhotoColor = val
End Property
Property Get PhotoColor() As String
    PhotoColor = vPhotoColor
End Property
Property Let PhotoIMG(val As String)
    vPhotoIMG = val
End Property
Property Get PhotoIMG() As String
    PhotoIMG = vPhotoIMG
End Property

Property Let PhotoRotate(val As Double)
    vPhotoRotate = val
End Property
Property Get PhotoRotate() As Double
    PhotoRotate = vPhotoRotate
End Property


Property Let Altitude(val As Long)
'IN METERS
    vAltitude = val
End Property
Property Get Altitude() As Long
'IN METERS
    Altitude = vAltitude
End Property
Property Let AltitudeMode(val As KMLAltitudeModes)
'FOR KML NAMESPACE ONLY, GX NOT SUPPORTED IN THIS DOC
'IF GX LOADED, CONVERTS TO KML NAMESPACE TYLES (KMLAltitudeModes)
    'clampToGround, relativeToGround, absolute
    vAltitudeMode = val
End Property
Property Get AltitudeMode() As KMLAltitudeModes
'Read after load only
    AltitudeMode = vAltitudeMode
End Property


Property Let Extended(val As Boolean)
'extrude the placemark to it's altitude
    vExtended = val
End Property
Property Get Extended() As Boolean
    Extended = vExtended
End Property







Sub Load(KMLString As String, Optional FailOnError As Boolean, Optional PreserveOnFail As Boolean)
'LOAD AN XML STRING INTO THE OBJECT
'PRESERVEONFAIL = DON'T CLEAR PROPERTIES IF FAILS
Dim xDoc            As New MSXML2.DOMDocument60
Dim xNode           As IXMLDOMNode
Dim xRoot           As IXMLDOMNode
Dim xChild          As IXMLDOMNode
Dim xGrandChild     As IXMLDOMNode
Dim xNodeList       As IXMLDOMNodeList
Dim xNodeList2      As IXMLDOMNodeList
Dim str             As String
Dim RootName        As String

If KMLString = "" Then
    AddDebug ".Load(): Failed - No string was supplied."
    GoTo handler
End If

On Error GoTo handler

ClearObject ' clear all properties
vDebugString = ""

KMLString = Trim(KMLString)

If InStr(1, KMLString, "gx:") > 0 Then
    KMLString = StripGXTag(KMLString)
    AddDebug ".Load():  GX namespace element(s) not supported in KML String. Stripping..."
End If

With xDoc
    .validateOnParse = True
    .resolveExternals = False
    .preserveWhiteSpace = True
    .async = False
    .loadXML (KMLString)
End With

If ValidateXML(KMLString) <> True Then
    AddDebug ".Load(): KML String provided is not valid XML. "
    GoTo handler
Else
    Set xNode = xDoc.selectSingleNode("/")
    Set xNodeList = xNode.childNodes
    
    If xNodeList.Length <> 1 Then
        AddDebug ".Load():  Failed to parse XML. Expected 1 root node."
        GoTo handler
    Else
        Set xRoot = xNodeList.Item(0)
        
        'name
        Set xChild = GetChildNode(xRoot, "name")
        If Not xChild Is Nothing Then Me.PlacemarkName = xChild.Text
        'descr
        Set xChild = GetChildNode(xRoot, "description")
        If Not xChild Is Nothing Then Me.Description = xChild.Text
        'style
        Set xChild = GetChildNode(xRoot, "styleUrl")
        If Not xChild Is Nothing Then Me.StyleURL = xChild.Text
        
        If xRoot.nodeName = "Placemark" Then
            'POINT
            Set xChild = GetChildNode(xRoot, "Point")
            If Not xChild Is Nothing Then
                Me.PlacemarkType = kPoint
                'coords
                Set xGrandChild = GetChildNode(xChild, "coordinates")
                If Not xGrandChild Is Nothing Then
                    If FormatCoords(xGrandChild.Text, True) = "" Then
                        AddDebug ".Load().FormatCoords() - Coordinates failed to format. Resuming..."
                    Else
                        Me.Coordinates = FormatCoords(xGrandChild.Text, True) 'strip altitude
                        Me.Altitude = GetCord(xGrandChild.Text, 1, 3)
                    End If
                End If
                'extrude
                Set xGrandChild = GetChildNode(xChild, "extrude")
                If Not xGrandChild Is Nothing Then
                    Me.Extended = CBool(xGrandChild.Text)
                End If
                
                'alt mode
                Set xGrandChild = GetChildNode(xChild, "altitudeMode")
                If Not xGrandChild Is Nothing Then
                    Me.AltitudeMode = ConvertAltitudeMode(xGrandChild.Text)
                End If
            End If
                
            'PATH - same as 'POINT' - nearly duplicate code
            Set xChild = GetChildNode(xRoot, "LineString")
            If Not xChild Is Nothing Then
                Me.PlacemarkType = kPath
                
                'coords
                Set xGrandChild = GetChildNode(xChild, "coordinates")
                If Not xGrandChild Is Nothing Then
                    If FormatCoords(xGrandChild.Text, True) = "" Then
                        AddDebug ".Load().FormatCoords() - Coordinates failed to format. Resuming..."
                    Else
                        Me.Coordinates = FormatCoords(xGrandChild.Text, True)
                        
                        Me.Altitude = GetCord(xGrandChild.Text, 1, 3)
                    End If
                End If
                'extrude
                Set xGrandChild = GetChildNode(xChild, "extrude")
                If Not xGrandChild Is Nothing Then
                    Me.Extended = CBool(xGrandChild.Text)
                End If
                
                'alt mode
                Set xGrandChild = GetChildNode(xChild, "altitudeMode")
                If Not xGrandChild Is Nothing Then
                    Me.AltitudeMode = ConvertAltitudeMode(xGrandChild.Text)
                End If
            End If
            
            'POLY
            Set xChild = GetChildNode(xRoot, "Polygon")
            If Not xChild Is Nothing Then
                Me.PlacemarkType = kPolygon
                
                'coords
                Set xGrandChild = GetChildNode(xChild, "outerBoundaryIs", "LinearRing")
                If Not xChild Is Nothing Then
                    Set xGrandChild = GetChildNode(xGrandChild, "coordinates")
                    If Not xGrandChild Is Nothing Then
                        If FormatCoords(xGrandChild.Text, True) = "" Then
                            AddDebug ".Load().FormatCoords() - Coordinates failed to format. Resuming..."
                        Else
                            Me.Coordinates = FormatCoords(xGrandChild.Text, True)
                            Me.Altitude = GetCord(xGrandChild.Text, 1, 3)
                        End If
                    End If
                End If
                    
                'extrude
                Set xGrandChild = GetChildNode(xChild, "extrude")
                If Not xGrandChild Is Nothing Then
                    Me.Extended = CBool(xGrandChild.Text)
                End If
                
                'alt mode
                Set xGrandChild = GetChildNode(xChild, "altitudeMode")
                If Not xGrandChild Is Nothing Then
                    Me.AltitudeMode = ConvertAltitudeMode(xGrandChild.Text)
                End If
            End If

        'PHOTO OVERLAY
        ElseIf xRoot.nodeName = "GroundOverlay" Then
            
            Me.PlacemarkType = kPhotoOverlay
            
            'IMAGE URL
            Set xChild = GetChildNode(xRoot, "Icon", "href")
            If Not xChild Is Nothing Then
                Me.PhotoIMG = xChild.Text
            End If
            
            'color
            Set xChild = GetChildNode(xRoot, "color")
            If Not xChild Is Nothing Then
                Me.PhotoColor = xChild.Text
            End If
            
            'altitude
            Set xChild = GetChildNode(xRoot, "altitude")
            If Not xChild Is Nothing Then
                Me.Altitude = xChild.Text
            End If
            
            'altitudemode
            Set xChild = GetChildNode(xRoot, "altitudeMode")
            If Not xChild Is Nothing Then
                Me.AltitudeMode = ConvertAltitudeMode(xChild.Text)
            End If
            
            'coordinates
            Set xChild = GetChildNode(xRoot, "LatLonBox", "north")
            If Not xChild Is Nothing Then
                str = xChild.Text
                'assume rest exist...
                str = str & " " & GetChildNode(xRoot, "LatLonBox", "south").Text
                str = str & " " & GetChildNode(xRoot, "LatLonBox", "east").Text
                str = str & " " & GetChildNode(xRoot, "LatLonBox", "west").Text
                Me.Coordinates = str
            
                'rotation
                Set xChild = GetChildNode(xRoot, "LatLonBox", "rotation")
                If Not xChild Is Nothing Then Me.PhotoRotate = CDbl(xChild.Text)
            End If
        End If
        
        'final check:
        If Me.PlacemarkType = 0 Then
            AddDebug ".Load() - A placemark type node was not found or is not supported."
            GoTo handler
        End If
    End If
End If

Exit Sub
handler:
If Err.Number <> vbObjectError + 513 And Err.Number <> 0 Then
    AddDebug ".Load() RUNTIME ERROR: " & Err.Number & " " & Err.Description
End If
If FailOnError = True Then Err.Raise vbObjectError + 513, , ".Load() - Failed to load KML string. See DebugString."
If PreserveOnFail = False Then ClearObject
End Sub



Function Build(Optional FailOnError As Boolean) As String
'BUILD XML STRING FOR THE PLACEMARK OBJECT
Dim output          As String
Dim Header          As String
Dim AltString       As String
Dim ExtString       As String
Dim TessString      As String
Dim Valid           As Variant

vDebugString = ""  'CLEAR
On Error GoTo handler

'DEFAULTS
If Me.PlacemarkType = 0 Then Me.PlacemarkType = 1 'point
If Me.PlacemarkName = "" Then Me.PlacemarkName = "Untitled placemark"
If Me.AltitudeMode = 0 Then Me.AltitudeMode = 1 'clamp to ground

'REQUIRED
If Me.Coordinates = "" Then
    AddDebug ".Build(): Failed - Coordinates cannot be empty."
    GoTo handler
End If

'FORCES
If Me.Altitude > 0 And Me.AltitudeMode = kClampToGround Then
    'note - this will be overriden again IF caller building Overlay and does not have absolute mode
    AddDebug ".Build(): Altitude was specified. Mode incorrect. Forcing 'relativeToGround'..."
    Me.AltitudeMode = kRelativeToGround
End If
If Me.Extended = True And Me.AltitudeMode = 1 Then
    AddDebug ".Build(): .Extended = TRUE not allowed with .AltitudeMode 'clampToGround'. Forcing 'False'..."
End If
If Me.StyleURL = "" Then Me.StyleURL = "EMPTY"

'STRINGS
Header = "<name>" & Me.PlacemarkName & _
              "</name><description>" & Me.Description & "</description>" & _
              "<styleUrl>" & Me.StyleURL & "</styleUrl>"
             
If Me.Altitude > 0 Then
    AltString = "<altitudeMode>" & _
                TranslateAltitudeMode(Me.AltitudeMode) & _
                "</altitudeMode>"
End If

If Me.Extended = True Then
    ExtString = "<extrude>1</extrude>"
End If

TessString = "<tessellate>1</tessellate>"

Select Case Me.PlacemarkType

    Case 1: 'point
        output = Header & "<Point>" & ExtString & AltString & "<coordinates>" & _
                 FormatCoords(Me.Coordinates) & "</coordinates></Point>"
        
    Case 2: 'path
    
        'check coords
        If FormatCoords(Me.Coordinates) = "" Then
            GoTo handler
        End If
            
    
        'note for tessellation - docs says must be 'clamptoground' or 'clamptoseafloor' but
        'earth allows 'clamptoground' or 'relativetoground' and NOT absoltue
        output = Header & "<LineString>" & ExtString & TessString & _
                AltString & "<coordinates>" & FormatCoords(Me.Coordinates) & _
                "</coordinates></LineString>"
                   
    Case 3:  'Polygon
        'check coords
            'check if end cord matches start cord

            If CountCords(Me.Coordinates) < 4 Then
                AddDebug ".Build(): Warning. Polygon should have at least 3 + 1 coordinates."
            ElseIf GetCord(Me.Coordinates, 1) <> _
                GetCord(Me.Coordinates, CountCords(Me.Coordinates)) Then
                AddDebug ".Build(): Warning. End cord for polygon should match start cord. Ignored..."
            End If
            
        output = Header & "<Polygon>" & ExtString & TessString & AltString & _
                 "<outerBoundaryIs><LinearRing><coordinates>" & _
                 FormatCoords(Me.Coordinates) & "</coordinates>" & _
                 "</LinearRing></outerBoundaryIs></Polygon>"
                'force tessellation - not used according to docs but test file has value of 1
                
            
    Case 4:  'Photo overlay
        If InStr(1, Me.Coordinates, ",") Then
            AddDebug ".Build(): Caller specified .PlacemarkType = 'GroundOverlay'. Coordinates improper format"
            GoTo handler
        End If
        
        If Me.Extended = True Then
            'extrude not supported with overlays
            AddDebug ".Build(): Caller specified 'Extended = True'. Ignoring..."
        End If
            
        If Me.Altitude > 0 Then
            If Me.AltitudeMode <> kAbsolute Then 'only absolute allowed with overlays with altitude
                Me.AltitudeMode = kAbsolute
                AddDebug ".Build(): Altitude was specified. Mode incorrect. Forcing 'absolute'..."
            End If
            'ESTABLISH NEW ALTSTRING
            'note that actual 'altitude' tag is included and not provided in coords...
            AltString = "<altitude>" & Me.Altitude & "</altitude><altitudeMode>" & _
            TranslateAltitudeMode(Me.AltitudeMode) & _
            "</altitudeMode>"
        Else
            AltString = ""
        End If
        
        output = Header & "<color>" & Me.PhotoColor & _
        "</color><Icon><href>" & Me.PhotoIMG & _
        "</href><viewBoundScale>0.75</viewBoundScale></Icon>" & _
        AltString
        
        'NSEW cords:
        output = output & "<LatLonBox><north>" & GetCord(Me.Coordinates, 1) & _
                 "</north><south>" & GetCord(Me.Coordinates, 2) & "</south>" & _
                 "<east>" & GetCord(Me.Coordinates, 3) & "</east>" & _
                 "<west>" & GetCord(Me.Coordinates, 4) & "</west>"
        If Me.PhotoRotate > 0 Then output = output & "<rotation>" & Me.PhotoRotate & "</rotation>"
        output = output & "</LatLonBox>"
            
End Select

If Me.PlacemarkType = kPhotoOverlay Then
    output = "<GroundOverlay>" & output & "</GroundOverlay>"
Else
    output = "<Placemark>" & output & "</Placemark>"
End If

'Format
Build = FormatXML(output)
'Validate output
Valid = ValidateXML(Build)
If Valid <> True Then
    AddDebug ".Build(): - " & Valid
    GoTo handler
End If

Exit Function
handler:
Build = ""
If Err.Number <> vbObjectError + 513 And Err.Number <> 0 Then
    AddDebug ".Build() RUNTIME ERROR: " & Err.Number & " " & Err.Description
End If
If FailOnError = True Then Err.Raise vbObjectError + 513, , ".Build() - Failed to build placemark. See DebugString."
End Function



Function FormatCoords(CoordString As String, Optional Deformat As Boolean = False) As String
'FORMATS/DEFORMATS A COORDINATES STRING
'FORMAT:  APPENDS ALTITUDE TO EACH CORDINATE IN STRING
'DEFORAMT: STRIPS ALTITUDE FROM EACH CORDINATE IN STRING
'ADDITIONALLY WILL INVERT THE COORDINATES BASED ON .InvertCoordinates property
Dim arr         As Variant
Dim arr2        As Variant

Dim i           As Integer
Dim k           As Integer
Dim output      As String
Dim ArrLen      As Integer

On Error GoTo handler
If Len(CoordString) = 0 Then Exit Function
CoordString = Replace(CoordString, vbLf, " ") 'an attempt to format a file which has coords separated by line breaks - not recommended
CoordString = Replace(CoordString, vbCrLf, " ")
CoordString = Replace(CoordString, vbTab, " ")
CoordString = TrimAll(CoordString)
CoordString = RemoveExtraSpaces(CoordString)

arr = Split(CoordString, " ")

For i = LBound(arr) To UBound(arr)
    arr2 = Split(CStr(arr(i)), ",")
    ArrLen = UBound(arr2) - LBound(arr2) + 1
    
    If Len(CStr(arr(i))) > 0 Then  'ignore blanks if present
    
        If Deformat = False Then
            If ArrLen = 2 Then
                If vInvertCoordinates = True Then
                    output = output & " " & arr2(1) & "," & arr2(0) & "," & Me.Altitude
                Else
                    output = output & " " & arr(i) & "," & Me.Altitude
                End If
            Else
                AddDebug ".FormatCoords():  Improper coordinate format"
                Exit Function
            End If
        Else
            If ArrLen = 3 Then
                If vInvertCoordinates = True Then
                    output = output & " " & arr2(1) & "," & arr2(0)
                Else
                    output = output & " " & arr2(0) & "," & arr2(1)
                End If
            Else
                AddDebug ".FormatCoords():  Improper coordinate format"
                Exit Function
            End If
        End If
    End If
Next i

output = Trim(output)
FormatCoords = output
Exit Function
handler:
    FormatCoords = ""
    AddDebug ".FormatCoords() - RUNTIME ERROR: " & Err.Number & "  " & Err.Description
    Err.Raise vbObjectError + 513  'let build routine handle error

End Function


'---------- PRIVATE -----------------------------------------

Private Function CountCords(CordString As String) As Integer
'return number of coords in cordinate string
CountCords = UBound(Split(CordString, " ")) - LBound(Split(CordString, " ")) + 1
End Function
Private Function RemoveExtraSpaces(CoordString As String) As String
'removes multiple spaces from a coordinate string
Dim word        As String
Dim nxtchr      As String

Dim arr         As Variant
Dim i           As Integer
Dim k           As Integer
Dim output      As String
Dim str As String

arr = Split(CoordString, " ")
For i = LBound(arr) To UBound(arr)
    word = arr(i)
    
    If word <> "" Then
        output = output & " " & word
    End If
Next i
RemoveExtraSpaces = Trim(output)
End Function

Private Function GetCord(CordString As String, index As Integer, Optional subindex As Integer) As String
'returns a coordinate from a set of coordinates
'index begins with 1
'subindex to return lat, lng, or altitude from coordinate
On Error Resume Next
If CordString = "" Then Exit Function
If InStr(1, CordString, ",") = 0 And InStr(1, CordString, " ") = 0 Then Exit Function

'attemp to format with spaces as delimeter
CordString = Replace(CordString, vbLf, " ")
CordString = Replace(CordString, vbCrLf, " ")
CordString = Replace(CordString, vbTab, " ")
CordString = TrimAll(CordString)
CordString = RemoveExtraSpaces(CordString)


GetCord = Trim(Split(CordString, " ")(index - 1))
If subindex > 0 Then GetCord = Trim(Split(GetCord, ",")(subindex - 1))
End Function

Private Function StripGXTag(KMLString As String) As String
'REMOVE GX PREFIX FROM KMLSTRING - not supported
Dim arr         As Variant
Dim i           As Integer
On Error GoTo handler
If InStr(1, KMLString, "gx:") > 0 Then
    arr = Split(KMLString, "<")
    
    For i = LBound(arr) To UBound(arr)
        If Left(arr(i), 3) = "gx:" Or Left(arr(i), 4) = "/gx:" Then
            arr(i) = Replace(CStr(arr(i)), "gx:", "")
            arr(i) = Replace(CStr(arr(i)), "/gx:", "")
        End If
    Next i
    StripGXTag = Join(arr, "<")
Else
    StripGXTag = KMLString
End If
Exit Function
handler:
    AddDebug ".StripGXTag():  Failed to strip GX tag from KMLString."
End Function


Private Function GetChildNode(ParentNode As IXMLDOMNode, ChildName As String, Optional GrandChildName As String) As IXMLDOMNode
'LOOKS FOR CHILD NODE OR GRANDCHILD NODE AND RETURNS NODE OBJECT
'IF NOT FOUND, RETURNS NOTHING
Dim xNode           As IXMLDOMNode
Dim xNode2          As IXMLDOMNode
Dim i               As Integer
On Error GoTo handler
Set GetChildNode = Nothing
For Each xNode In ParentNode.childNodes
    If xNode.nodeName = ChildName Then
        If GrandChildName <> "" Then
            For Each xNode2 In xNode.childNodes
                If xNode2.nodeName = GrandChildName Then
                    Set GetChildNode = xNode2
                    Exit Function
                End If
            Next xNode2
            Exit Function
        End If
        Set GetChildNode = xNode
        Exit Function
    End If
Next xNode
Exit Function
handler:
    AddDebug ".GetChildNode(): RUNTIME ERROR: " & Err.Number & " " & Err.Description
End Function


Private Function TranslateAltitudeMode(EnumVal As Integer) As String
'translates enum value to KML string
Select Case EnumVal
    Case 1: TranslateAltitudeMode = "clampToGround"
    Case 2: TranslateAltitudeMode = "relativeToGround"
    Case 3:  TranslateAltitudeMode = "absolute"
End Select
End Function

Private Function ConvertAltitudeMode(AltModeString As String) As KMLAltitudeModes
'BACKWARDS TRANSLATES KML Altitude mode string to object enum value
'ALSO HANDLES 'GX' NAMESPACE AND CONVERTS TO APPROPRIATE KML NAMESPACE MODE

Select Case AltModeString
    Case "clampToGround": ConvertAltitudeMode = kClampToGround
    Case "relativeToGround": ConvertAltitudeMode = kRelativeToGround
    Case "absolute": ConvertAltitudeMode = kAbsolute
    Case "relativeToSeaFloor": ConvertAltitudeMode = kRelativeToGround  'gx
    Case "clampToSeaFloor": ConvertAltitudeMode = kClampToGround        'gx
End Select
End Function

Private Sub ClearObject()
    Me.Altitude = 0
    Me.AltitudeMode = 0
    Me.Coordinates = ""
    Me.Description = ""
    Me.Extended = False
    Me.PhotoColor = 0
    Me.PhotoIMG = ""
    Me.PhotoRotate = 0
    Me.PlacemarkName = ""
    Me.PlacemarkType = 0
    Me.StyleURL = ""
End Sub
Private Sub AddDebug(Message As String)
'log to debug string
Dim Header As String
vDebugString = vDebugString & vbCrLf & "DEBUG: .PlacemarkName = '" & Me.PlacemarkName & "'" & vbTab & Message
End Sub

Private Function ValidateXML(XMLString As String) As Variant
Dim xDoc            As MSXML2.DOMDocument60
On Error GoTo handler
Set xDoc = New MSXML2.DOMDocument60
With xDoc
    .validateOnParse = True
    .resolveExternals = False
    .preserveWhiteSpace = True
    .async = False
    
    .loadXML XMLString
End With
With xDoc.parseError
If .reason = "" Then
    ValidateXML = True
Else
    ValidateXML = "InvalidXML  " & .errorCode & "-  " & .reason
End If
End With
Exit Function
handler:
    AddDebug ".ValidateXML() - RUNTIME ERROR: " & Err.Number & "  " & Err.Description
    Err.Raise vbObjectError + 513  'let build routine handle error
End Function

Private Function FormatXML(XMLString As String) As String
'https://stackoverflow.com/questions/1118576/how-can-i-pretty-print-xml-source-using-vb6-and-msxml
  Dim Reader As New SAXXMLReader60
  Dim Writer As New MXXMLWriter60
  
On Error GoTo handler
  Writer.indent = True
  Writer.standalone = False

  Writer.omitXMLDeclaration = False

  Writer.Encoding = "utf-8"

  Set Reader.contentHandler = Writer
  Set Reader.dtdHandler = Writer
  Set Reader.errorHandler = Writer

  Call Reader.putProperty("http://xml.org/sax/properties/declaration-handler", _
          Writer)
  Call Reader.putProperty("http://xml.org/sax/properties/lexical-handler", _
          Writer)

  Call Reader.parse(XMLString)

  FormatXML = Writer.output
  'remove standalone declaration:
  FormatXML = Replace(FormatXML, " standalone=" & Chr(34) & "no" & Chr(34) & "?>" & vbCrLf, "?>")
  
  'remove XML declaration only if not present in XMLstring
  If InStr(1, XMLString, "<?xml version=" & Chr(34) & "1.0" & Chr(34) & "?>") = 0 Then
    FormatXML = Replace(FormatXML, "<?xml version=" & Chr(34) & "1.0" & Chr(34) & "?>", "")
  End If
  Exit Function
handler:
    AddDebug ".FormatXML() - RUNTIME ERROR: " & Err.Number & "  " & Err.Description
    Err.Raise vbObjectError + 513  'let build routine handle error
End Function




Private Function TrimAll(Text As String) As String
'trims all whitespace characters
'https://stackoverflow.com/questions/25184019/trim-all-types-of-whitespace-including-tabs

On Error Resume Next

Dim toRemove As String
toRemove = " " & vbTab & vbCr & vbLf & vbCrLf & Chr(13) & Chr(9) & Chr(10) 'what to remove

Dim s As Long: s = 1
Dim e As Long: e = Len(Text)
Dim c As String

If e = 0 Then Exit Function 'zero len string

Do 'how many chars to skip on the left side
    c = Mid(Text, s, 1)
    If c = "" Or InStr(1, toRemove, c) = 0 Then Exit Do
    s = s + 1
Loop
    
Do 'how many chars to skip on the right side
    c = Mid(Text, e, 1)
    If e = 1 Or InStr(1, toRemove, c) = 0 Then Exit Do
    e = e - 1
Loop
TrimAll = Mid(Text, s, (e - s) + 1) 'return remaining text
End Function



































